from typing import Any, Optional, Type

from pydantic import BaseModel, validator

from phidata.utils.log import logger


class InfraResource(BaseModel):
    """Base class for all Phidata infrastructure resources.
    All Models in the phidata.infra.*.resource modules are expected to be subclasses of this Model.

    We use a pydantic model for resources because the data which creates the resource
    comes from an external sources  like users or the api.
    This data needs to be validated & type checked for which we use pydantic.
    """

    # name of resource
    name: Optional[str] = None
    # type of resource
    resource_type: Optional[str] = None

    # resource management
    # If True, skip resource creation if an active resources with the same name exists.
    use_cache: bool = True

    # If True, marks skip_create, skip_delete, skip_update to True
    external: bool = False
    # If True, does not create the resource when `phi ws up` command is run
    skip_create: bool = False
    # If True, does not delete the resource when `phi ws down` command is run
    skip_delete: bool = False
    # If True, does not update the resource when `phi ws patch` command is run
    skip_update: bool = False

    # If True, waits for the resource to be created
    wait_for_creation: bool = True
    # If True, waits for the resource to be deleted
    wait_for_deletion: bool = True
    # The amount of time in seconds to wait between attempts.
    waiter_delay: int = 30
    # The maximum number of attempts to be made.
    waiter_max_attempts: int = 50

    active_resource: Optional[Any] = None
    active_resource_class: Optional[Type] = None

    def get_resource_name(self) -> Optional[str]:
        return self.name

    def get_resource_type(self) -> Optional[str]:
        return self.resource_type

    @validator("skip_create", pre=True, always=True)
    def set_skip_create(cls, skip_create, values):
        external = values.get("external", False)
        return True if external else skip_create

    @validator("skip_delete", pre=True, always=True)
    def set_skip_delete(cls, skip_delete, values):
        external = values.get("external", False)
        return True if external else skip_delete

    @validator("skip_update", pre=True, always=True)
    def set_skip_update(cls, skip_update, values):
        external = values.get("external", False)
        return True if external else skip_update

    """
    ## Functions to be implemented by subclasses
    def create(self, infra_api_client: InfraApiClient) -> bool:
    def read(self, infra_api_client: InfraApiClient) -> bool:
    def update(self, infra_api_client: InfraApiClient) -> bool:
    def delete(self, infra_api_client: InfraApiClient) -> bool:
    def is_active(self, infra_api_client: InfraApiClient) -> bool:
    """

    class Config:
        # https://pydantic-docs.helpmanual.io/usage/model_config/
        # If we need to use an alias for fields of subclasses, eg: Kubeconfig
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        use_enum_values = True



from typing import Optional, Union, Sequence
from typing_extensions import Literal

from sqlalchemy.engine import Engine, Connection

from phidata.asset.table.sql import SqlTable, SqlTableArgs, SqlType
from phidata.utils.cli_console import print_info, print_error


class PostgresTableArgs(SqlTableArgs):
    # Table Name
    name: str
    sql_type: SqlType = SqlType.POSTGRES


class PostgresTable(SqlTable):
    def __init__(
        self,
        # Table Name
        name: str,
        db_schema: Optional[str] = None,
        db_conn_id: Optional[str] = None,
        db_conn_url: Optional[str] = None,
        db_engine: Optional[Union[Engine, Connection]] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:
        super().__init__()
        try:
            self.args: PostgresTableArgs = PostgresTableArgs(
                name=name,
                db_conn_id=db_conn_id,
                db_conn_url=db_conn_url,
                db_engine=db_engine,
                db_schema=db_schema,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            print_error(f"Args for {self.__class__.__name__} are not valid")
            raise



from pathlib import Path
from typing import Optional, Dict, Union

from phidata.base import PhidataBaseArgs


class InfraArgs(PhidataBaseArgs):
    env: Optional[str] = None

    # -*- Path parameters
    # Path to the workspace root directory
    workspace_root_path: Optional[Path] = None
    # Path to the workspace config file
    workspace_config_file_path: Optional[Path] = None
    # Path to important directories relative to the workspace_root
    # These directories are joined to the workspace_root dir
    #   to build paths depending on the environments (local, docker, k8s)
    # defaults are set by WorkspaceConfig.__init__()
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    meta_dir: Optional[str] = None
    products_dir: Optional[str] = None
    notebooks_dir: Optional[str] = None
    workspace_config_dir: Optional[str] = None

    # -*- Environment parameters
    # Env vars added to local env
    local_env: Optional[Dict[str, str]] = None
    # Load local env using env file
    local_env_file: Optional[Union[str, Path]] = None
    # Env vars added to docker env
    docker_env: Optional[Dict[str, str]] = None
    # Load docker env using env file
    docker_env_file: Optional[Union[str, Path]] = None
    # Env vars added to k8s env
    k8s_env: Optional[Dict[str, str]] = None
    # Load k8s env using env file
    k8s_env_file: Optional[Union[str, Path]] = None

    # -*- AWS parameters
    # Common aws params used by apps, resources and data assets
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None



from pathlib import Path
from typing import Optional, Dict, Any
from typing_extensions import Literal

from phidata.app.db import DbApp
from phidata.app.superset.superset_base import SupersetBase
from phidata.infra.k8s.enums.restart_policy import RestartPolicy


class SupersetManager(SupersetBase):
    def __init__(
        self,
        name: str = "superset-manager",
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "phidata/superset",
        image_tag: str = "2.3.0",
        entrypoint: str = "/manager.sh",
        command: Optional[str] = None,
        # Creates an superset admin with username: admin, pass: admin
        # or reads details from secrets file
        create_superset_admin_user: bool = False,
        # Configure superset db,
        # If init_superset_db = True, initialize the superset_db,
        init_superset_db: bool = False,
        # Upgrade the superset db
        upgrade_superset_db: bool = False,
        wait_for_db: bool = True,
        # delay start by 60 seconds for the db to be initialized
        wait_for_db_init: bool = False,
        # Connect to database using DbApp,
        db_app: Optional[DbApp] = None,
        # Connect to database manually,
        db_user: Optional[str] = None,
        db_password: Optional[str] = None,
        db_schema: Optional[str] = None,
        db_host: Optional[str] = None,
        db_port: Optional[int] = None,
        db_driver: str = "postgresql+psycopg2",
        db_result_backend_driver: str = "db+postgresql",
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the deployment,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Add deployment labels
        deploy_labels: Optional[Dict[str, Any]] = None,
        # Determine how to spread the deployment across a topology
        # Key to spread the pods across
        topology_spread_key: Optional[str] = None,
        # The degree to which pods may be unevenly distributed
        topology_spread_max_skew: Optional[int] = None,
        # How to deal with a pod if it doesn't satisfy the spread constraint.
        topology_spread_when_unsatisfiable: Optional[
            Literal["DoNotSchedule", "ScheduleAnyway"]
        ] = None,
        print_env_on_load: bool = True,
        # Additional args
        # If True, use cached resources
        # i.e. skip resource creation/deletion if active resources with the same name exist.
        use_cache: bool = True,
    ):
        super().__init__(
            name=name,
            version=version,
            enabled=enabled,
            image_name=image_name,
            image_tag=image_tag,
            entrypoint=entrypoint,
            command=command,
            create_superset_admin_user=create_superset_admin_user,
            init_superset_db=init_superset_db,
            upgrade_superset_db=upgrade_superset_db,
            wait_for_db=wait_for_db,
            wait_for_db_init=wait_for_db_init,
            db_app=db_app,
            db_user=db_user,
            db_password=db_password,
            db_schema=db_schema,
            db_host=db_host,
            db_port=db_port,
            db_driver=db_driver,
            db_result_backend_driver=db_result_backend_driver,
            env=env,
            env_file=env_file,
            secrets_file=secrets_file,
            deploy_name=deploy_name,
            pod_name=pod_name,
            replicas=replicas,
            pod_node_selector=pod_node_selector,
            restart_policy=restart_policy,
            termination_grace_period_seconds=termination_grace_period_seconds,
            deploy_labels=deploy_labels,
            topology_spread_key=topology_spread_key,
            topology_spread_max_skew=topology_spread_max_skew,
            topology_spread_when_unsatisfiable=topology_spread_when_unsatisfiable,
            print_env_on_load=print_env_on_load,
            use_cache=use_cache,
        )



# import datetime
# from typing import Any, Dict, List, Optional
#
# import google.auth.transport.requests as google_auth_requests
# from google.cloud import container_v1
# from google.oauth2 import service_account
#
# import phidata.infra.k8s.enums as k8s_enums
# from phidata.infra.k8s.resource.kubeconfig import (
#     KubeconfigCluster,
#     KubeconfigUser,
#     KubeconfigUserConfig,
#     KubeconfigAuthConfig,
#     KubeconfigAuthProvider,
#     KubeconfigContext,
#     KubeconfigContextSpec,
#     Kubeconfig,
#     KubeconfigClusterConfig,
# )
# from phidata.utils.log import logger
#
#
# def _create_kubeconfig_cluster(
#     cluster_name: str, cluster_endpoint: str, cluster_ca_certificate: str
# ) -> KubeconfigCluster:
#     _cluster = KubeconfigCluster(
#         name=cluster_name,
#         cluster=KubeconfigClusterConfig(
#             server="https://{}".format(cluster_endpoint),
#             certificate_authority_data=cluster_ca_certificate,
#             # "insecure-skip-tls-verify": True,
#         ),
#     )
#     return _cluster
#
#
# def _create_kubeconfig_user(
#     auth_provider_name: str,
#     user_name: str,
#     access_token: str,
#     expiry: datetime.datetime,
# ) -> KubeconfigUser:
#     _user = KubeconfigUser(
#         name=user_name,
#         user=KubeconfigUserConfig(
#             auth_provider=KubeconfigAuthProvider(
#                 name=auth_provider_name,
#                 config=KubeconfigAuthConfig(
#                     access_token=access_token,
#                     expiry=expiry,
#                 ),
#             )
#         ),
#     )
#     return _user
#
#
# def _create_kubeconfig_context(
#     context_name: str, user_name: str, cluster_name: str, namespace: str
# ) -> KubeconfigContext:
#     _context = KubeconfigContext(
#         name=context_name,
#         context=KubeconfigContextSpec(
#             cluster=cluster_name,
#             user=user_name,
#             namespace=namespace,
#         ),
#     )
#     return _context
#
#
# def create_kubeconfig_resource_for_gke_cluster(
#     gke_cluster_gcp: container_v1.types.Cluster,
#     credentials: service_account.Credentials,
#     namespace_name: str,
#     context_name: str,
# ) -> Optional[Kubeconfig]:
#     """Creates the kubeconfig for a given cluster.
#
#     TODO:
#         * Learn more about kubeconfig - no idea what the fuck im doing here
#         * Figure out if there is a better way to generate a kubeconfig
#         * Add doc
#         * Add Error Handling
#         * Change input from container_v1.types.Cluster to resource.GKEClusterOutput
#     References:
#         * https://kubernetes.io/docs/concepts/configuration/organize-cluster-access-kubeconfig/
#         * https://kubernetes.io/docs/tasks/access-application-cluster/configure-access-multiple-clusters/
#         * https://godoc.org/k8s.io/client-go/tools/clientcmd/api#Config
#         * https://unofficial-kubernetes.readthedocs.io/en/latest/concepts/cluster-administration/authenticate-across-clusters-kubeconfig/
#     """
#
#     kubeconfig: Optional[Kubeconfig] = None
#     if gke_cluster_gcp and credentials:
#         try:
#             cluster_name = gke_cluster_gcp.name
#         except Exception as e:
#             logger.exception(f"Cannot parse GKEClusterSchema {gke_cluster_gcp}", e)
#             return None
#
#         logger.debug(f"Creating kubeconfig for {cluster_name}")
#         http_request = google_auth_requests.Request()
#         logger.debug("Refreshing credentials")
#         credentials.refresh(http_request)
#         logger.debug("Credentials refreshed")
#
#         # logger.debug(f"service_account_email: {credentials.service_account_email}")
#         # logger.debug(f"project_id: {credentials.project_id}")
#         # logger.debug(f"requires_scopes: {credentials.requires_scopes}")
#         # logger.debug(f"signer: {credentials.signer}")
#         # logger.debug(f"signer_email: {credentials.signer_email}")
#         # logger.debug(f"valid: {credentials.valid}")
#         # logger.debug(f"expired: {credentials.expired}")
#         # logger.debug(f"token: {credentials.token}")
#         # logger.debug(f"expiry: {credentials.expiry}")
#         # logger.debug(f"expiry type: {type(credentials.expiry)}")
#         # logger.debug(f"_credentials: {credentials}")
#         # logger.debug(f"gke_cluster_gcp: {gke_cluster_gcp}")
#
#         try:
#             # Cluster details for the kubeconfig
#             cluster_endpoint = gke_cluster_gcp.endpoint
#             cluster_ca_certificate = gke_cluster_gcp.master_auth.cluster_ca_certificate
#
#             # UserSchema details for the kubeconfig
#             service_account_email = credentials.service_account_email
#             access_token = credentials.token
#             expiry = credentials.expiry
#         except Exception as e:
#             logger.exception(f"Cannot Parse GKEClusterSchema {gke_cluster_gcp}", e)
#             return None
#
#         kubeconfig_cluster: KubeconfigCluster = _create_kubeconfig_cluster(
#             cluster_name, cluster_endpoint, cluster_ca_certificate
#         )
#         kubeconfig_user: KubeconfigUser = _create_kubeconfig_user(
#             "gcp", service_account_email, access_token, expiry
#         )
#         kubeconfig_context: KubeconfigContext = _create_kubeconfig_context(
#             context_name, service_account_email, cluster_name, namespace_name
#         )
#
#         kubeconfig = Kubeconfig(
#             api_version=k8s_enums.ApiVersion.CORE_V1,
#             kind=k8s_enums.Kind.CONFIG,
#             clusters=[kubeconfig_cluster],
#             users=[kubeconfig_user],
#             contexts=[kubeconfig_context],
#             current_context=context_name,
#         )
#
#     return kubeconfig



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Any, List, Union
from typing_extensions import Literal

from phidata.app.db import DbApp
from phidata.app import PhidataApp, PhidataAppArgs
from phidata.constants import (
    SCRIPTS_DIR_ENV_VAR,
    STORAGE_DIR_ENV_VAR,
    META_DIR_ENV_VAR,
    PRODUCTS_DIR_ENV_VAR,
    NOTEBOOKS_DIR_ENV_VAR,
    WORKSPACE_CONFIG_DIR_ENV_VAR,
    PHIDATA_RUNTIME_ENV_VAR,
)
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment, RestartPolicy
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.service import CreateService, ServiceType
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer, ImagePullPolicy
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    HostPathVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_service_name,
    get_default_deploy_name,
    get_default_pod_name,
    get_default_volume_name,
)
from phidata.utils.cli_console import print_error, print_warning
from phidata.utils.log import logger


class SupersetBaseArgs(PhidataAppArgs):
    name: str = "superset"
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "apache/superset"
    image_tag: str = "latest"
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # Mount the workspace directory on the container
    mount_workspace: bool = False
    workspace_volume_name: Optional[str] = None
    # Path to mount the workspace volume under
    # This is the parent directory for the workspace on the container
    # i.e. the ws is mounted as a subdir in this dir
    # eg: if ws name is: idata, workspace_dir would be: /usr/local/idata
    workspace_parent_container_path: str = "/usr/local"
    # NOTE: On DockerContainers the workspace_root_path is mounted to workspace_dir
    # because we assume that DockerContainers are running locally on the user's machine
    # On K8sContainers, we load the workspace_dir from git using a git-sync sidecar container
    create_git_sync_sidecar: bool = True
    git_sync_repo: Optional[str] = None
    git_sync_branch: Optional[str] = None
    git_sync_wait: int = 1
    # But when running k8s locally, we can mount the workspace using
    # host path as well.
    k8s_mount_local_workspace: bool = False

    # Superset resources directory relative to the workspace_root
    # This directory contains all the files required by superset.
    # eg: docker-bootstrap.sh
    # This dir is mounted to the `/app/docker` directory on the container
    resources_dir: str = "superset"
    resources_dir_container_path: str = "/app/docker"
    mount_resources: bool = True
    resources_volume_name: Optional[str] = None
    # skips downloading resources if the resources_dir exists
    cache_resources: bool = True

    # Set the SUPERSET_CONFIG_PATH env var
    superset_config_path: Optional[str] = None

    # Set the REQUIREMENTS_LOCAL env var
    # defaults to "/app/docker/requirements-local.txt"
    requirements_local: Optional[str] = None

    # Set the PYTHONPATH env var
    # defaults to "/app/pythonpath"
    python_path: Optional[str] = None

    # Configure Superset database
    # Get database details using DbApp
    db_app: Optional[DbApp] = None
    # Provide database details
    # Set the DATABASE_USER env var
    db_user: str = "superset"
    # Set the DATABASE_PASSWORD env var
    db_password: str = "superset"
    # Set the DATABASE_DB env var
    db_schema: str = "superset"
    # Set the DATABASE_HOST env var
    db_host: str = "db"
    # Set the DATABASE_PORT env var
    db_port: int = 5432
    # Set the DATABASE_DIALECT env var
    db_dialect: str = "postgresql+psycopg2"
    # Superset db connections in the format { conn_id: conn_url }
    db_connections: Optional[Dict] = None

    # Configure superset redis
    # Get redis details using PhidataApp
    redis_app: Optional[Any] = None
    # Provide redis details
    # Set the REDIS_HOST env var
    redis_host: str = "redis"
    # Set the REDIS_PORT env var
    redis_port: int = 6379

    # Set the FLASK_ENV env var
    flask_env: str = "production"
    # Set the SUPERSET_ENV env var
    superset_env: str = "production"
    # Set the SUPERSET_LOAD_EXAMPLES env var
    superset_load_examples: str = "yes"

    # Configure the container
    container_name: Optional[str] = None
    image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT
    container_detach: bool = True
    container_auto_remove: bool = True
    container_remove: bool = True

    # Add container labels
    container_labels: Optional[Dict[str, Any]] = None
    # NOTE: Available only for Docker
    # Add volumes to DockerContainer
    # container_volumes is a dictionary which adds the volumes to mount
    # inside the container. The key is either the host path or a volume name,
    # and the value is a dictionary with 2 keys:
    #   bind - The path to mount the volume inside the container
    #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    container_volumes: Optional[Dict[str, dict]] = None

    # Open a container port if open_container_port=True
    open_container_port: bool = False
    # Port number on the container
    container_port: int = 8000
    # Port name: Only used by the K8sContainer
    container_port_name: str = "http"
    # Host port: Only used by the DockerContainer
    container_host_port: int = 8000

    # Open the app port if open_app_port=True
    open_app_port: bool = True
    # App port number on the container
    # Set the SUPERSET_PORT env var
    app_port: int = 8088
    # Only used by the K8sContainer
    app_port_name: str = "app"
    # Only used by the DockerContainer
    app_host_port: int = 8088

    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Configure the ConfigMap used for env variables that are not Secret
    config_map_name: Optional[str] = None
    # Configure the Secret used for env variables that are Secret
    secret_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # Configure the deployment
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None
    replicas: int = 1
    pod_node_selector: Optional[Dict[str, str]] = None
    restart_policy: RestartPolicy = RestartPolicy.ALWAYS
    termination_grace_period_seconds: Optional[int] = None
    # Add deployment labels
    deploy_labels: Optional[Dict[str, Any]] = None
    # Determine how to spread the deployment across a topology
    # Key to spread the pods across
    topology_spread_key: Optional[str] = None
    # The degree to which pods may be unevenly distributed
    topology_spread_max_skew: Optional[int] = None
    # How to deal with a pod if it doesn't satisfy the spread constraint.
    topology_spread_when_unsatisfiable: Optional[
        Literal["DoNotSchedule", "ScheduleAnyway"]
    ] = None

    # Configure the app service
    create_app_service: bool = False
    app_service_name: Optional[str] = None
    app_service_type: Optional[ServiceType] = None
    # The port that will be exposed by the service.
    app_service_port: int = 8088
    # The node_port that will be exposed by the service if app_service_type = ServiceType.NODE_PORT
    app_node_port: Optional[int] = None
    # The app_target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    app_target_port: Optional[Union[str, int]] = None
    # Add labels to app service
    app_service_labels: Optional[Dict[str, Any]] = None


class SupersetBase(PhidataApp):
    def __init__(
        self,
        name: str = "superset",
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "apache/superset",
        image_tag: str = "latest",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Mount the workspace directory on the container,
        mount_workspace: bool = False,
        workspace_volume_name: Optional[str] = None,
        # Path to mount the workspace volume under,
        # This is the parent directory for the workspace on the container,
        # i.e. the ws is mounted as a subdir in this dir,
        # eg: if ws name is: idata, workspace_dir would be: /usr/local/idata,
        workspace_parent_container_path: str = "/usr/local",
        # NOTE: On DockerContainers the workspace_root_path is mounted to workspace_dir,
        # because we assume that DockerContainers are running locally on the user's machine,
        # On K8sContainers, we load the workspace_dir from git using a git-sync sidecar container,
        create_git_sync_sidecar: bool = True,
        git_sync_repo: Optional[str] = None,
        git_sync_branch: Optional[str] = None,
        git_sync_wait: int = 1,
        # But when running k8s locally, we can mount the workspace using,
        # host path as well.,
        k8s_mount_local_workspace: bool = False,
        # Superset resources directory relative to the workspace_root,
        # This directory contains all the files required by superset.,
        # eg: docker-bootstrap.sh,
        # This dir is mounted to the `/app/docker` directory on the container,
        resources_dir: str = "superset",
        resources_dir_container_path: str = "/app/docker",
        mount_resources: bool = True,
        resources_volume_name: Optional[str] = None,
        # skips downloading resources if the resources_dir exists,
        cache_resources: bool = True,
        # Set the SUPERSET_CONFIG_PATH env var,,
        superset_config_path: Optional[str] = None,
        # Set the REQUIREMENTS_LOCAL env var,
        # defaults to "/app/docker/requirements-local.txt",
        requirements_local: Optional[str] = None,
        # Set the PYTHONPATH env var,
        # defaults to "/app/pythonpath",
        python_path: Optional[str] = None,
        # Configure Superset database,
        # Get database details using DbApp,
        db_app: Optional[DbApp] = None,
        # Provide database details,
        # Set the DATABASE_USER env var,
        db_user: Optional[str] = None,
        # Set the DATABASE_PASSWORD env var,
        db_password: Optional[str] = None,
        # Set the DATABASE_DB env var,
        db_schema: Optional[str] = None,
        # Set the DATABASE_HOST env var,
        db_host: Optional[str] = None,
        # Set the DATABASE_PORT env var,
        db_port: Optional[int] = None,
        # Set the DATABASE_DIALECT env var,
        db_dialect: str = "postgresql+psycopg2",
        # Superset db connections in the format { conn_id: conn_url },
        db_connections: Optional[Dict] = None,
        # Configure superset redis,
        # Get redis details using PhidataApp,
        redis_app: Optional[Any] = None,
        # Provide redis details,
        # Set the REDIS_HOST env var,
        redis_host: str = "redis",
        # Set the REDIS_PORT env var,
        redis_port: int = 6379,
        # Set the FLASK_ENV env var,
        flask_env: str = "production",
        # Set the SUPERSET_ENV env var,
        superset_env: str = "production",
        # Set the SUPERSET_LOAD_EXAMPLES env var,
        superset_load_examples: str = "yes",
        # Configure the container,
        container_name: Optional[str] = None,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Add container labels,
        container_labels: Optional[Dict[str, Any]] = None,
        # NOTE: Available only for Docker,
        # Add volumes to DockerContainer,
        # container_volumes is a dictionary which adds the volumes to mount,
        # inside the container. The key is either the host path or a volume name,,
        # and the value is a dictionary with 2 keys:,
        #   bind - The path to mount the volume inside the container,
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.,
        # For example:,
        # {,
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},,
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'},
        # },
        container_volumes: Optional[Dict[str, dict]] = None,
        # Open a container port if open_container_port=True,
        open_container_port: bool = False,
        # Port number on the container,
        container_port: int = 8000,
        # Port name: Only used by the K8sContainer,
        container_port_name: str = "http",
        # Host port: Only used by the DockerContainer,
        container_host_port: int = 8000,
        # Open the app port if open_app_port=True,
        open_app_port: bool = False,
        # App port number on the container,
        app_port: int = 8088,
        # Only used by the K8sContainer,
        app_port_name: str = "app",
        # Only used by the DockerContainer,
        app_host_port: int = 8088,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the deployment,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Add deployment labels,
        deploy_labels: Optional[Dict[str, Any]] = None,
        # Determine how to spread the deployment across a topology,
        # Key to spread the pods across,
        topology_spread_key: Optional[str] = None,
        # The degree to which pods may be unevenly distributed,
        topology_spread_max_skew: Optional[int] = None,
        # How to deal with a pod if it doesn't satisfy the spread constraint.,
        topology_spread_when_unsatisfiable: Optional[
            Literal["DoNotSchedule", "ScheduleAnyway"],
        ] = None,
        # Configure the app service,
        create_app_service: bool = False,
        app_service_name: Optional[str] = None,
        app_service_type: Optional[ServiceType] = None,
        # The port that will be exposed by the service.,
        app_service_port: int = 8088,
        # The node_port that will be exposed by the service if app_service_type = ServiceType.NODE_PORT,
        app_node_port: Optional[int] = None,
        # The app_target_port is the port to access on the pods targeted by the service.,
        # It can be the port number or port name on the pod.,
        app_target_port: Optional[Union[str, int]] = None,
        # Add labels to app service,
        app_service_labels: Optional[Dict[str, Any]] = None,
        # Additional args
        # If True, use cached resources
        # i.e. skip resource creation/deletion if active resources with the same name exist.
        use_cache: bool = True,
    ):
        super().__init__()
        try:
            self.args: SupersetBaseArgs = SupersetBaseArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_parent_container_path=workspace_parent_container_path,
                create_git_sync_sidecar=create_git_sync_sidecar,
                git_sync_repo=git_sync_repo,
                git_sync_branch=git_sync_branch,
                git_sync_wait=git_sync_wait,
                k8s_mount_local_workspace=k8s_mount_local_workspace,
                resources_dir=resources_dir,
                resources_dir_container_path=resources_dir_container_path,
                mount_resources=mount_resources,
                resources_volume_name=resources_volume_name,
                cache_resources=cache_resources,
                superset_config_path=superset_config_path,
                requirements_local=requirements_local,
                python_path=python_path,
                db_app=db_app,
                db_user=db_user,
                db_password=db_password,
                db_schema=db_schema,
                db_host=db_host,
                db_port=db_port,
                db_dialect=db_dialect,
                db_connections=db_connections,
                redis_app=redis_app,
                redis_host=redis_host,
                redis_port=redis_port,
                flask_env=flask_env,
                superset_env=superset_env,
                superset_load_examples=superset_load_examples,
                container_name=container_name,
                image_pull_policy=image_pull_policy,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                container_labels=container_labels,
                container_volumes=container_volumes,
                open_container_port=open_container_port,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                open_app_port=open_app_port,
                app_port=app_port,
                app_port_name=app_port_name,
                app_host_port=app_host_port,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                deploy_name=deploy_name,
                pod_name=pod_name,
                replicas=replicas,
                pod_node_selector=pod_node_selector,
                restart_policy=restart_policy,
                termination_grace_period_seconds=termination_grace_period_seconds,
                deploy_labels=deploy_labels,
                topology_spread_key=topology_spread_key,
                topology_spread_max_skew=topology_spread_max_skew,
                topology_spread_when_unsatisfiable=topology_spread_when_unsatisfiable,
                create_app_service=create_app_service,
                app_service_name=app_service_name,
                app_service_type=app_service_type,
                app_service_port=app_service_port,
                app_node_port=app_node_port,
                app_target_port=app_target_port,
                app_service_labels=app_service_labels,
                use_cache=use_cache,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_app_service_name(self) -> str:
        return self.args.app_service_name or get_default_service_name(self.args.name)

    def get_app_service_port(self) -> int:
        return self.args.app_service_port

    def get_env_data_from_file(self) -> Optional[Dict[str, str]]:
        env_file_path = self.args.env_file
        if (
            env_file_path is not None
            and env_file_path.exists()
            and env_file_path.is_file()
        ):
            if env_file_path.suffix == ".yaml":
                import yaml

                # logger.debug(f"Reading {env_file_path}")
                env_data_from_file = yaml.safe_load(env_file_path.read_text())
                if env_data_from_file is not None and isinstance(
                    env_data_from_file, dict
                ):
                    return env_data_from_file
                else:
                    print_error(f"Invalid env_file: {env_file_path}")
            else:
                print_warning(f"Skipping: {env_file_path}")
        return None

    def get_secret_data_from_file(self) -> Optional[Dict[str, str]]:
        secrets_file_path = self.args.secrets_file
        if (
            secrets_file_path is not None
            and secrets_file_path.exists()
            and secrets_file_path.is_file()
        ):
            if secrets_file_path.suffix == ".yaml":
                import yaml

                # logger.debug(f"Reading {secrets_file_path}")
                secret_data_from_file = yaml.safe_load(secrets_file_path.read_text())
                if secret_data_from_file is not None and isinstance(
                    secret_data_from_file, dict
                ):
                    return secret_data_from_file
                else:
                    print_error(f"Invalid secrets_file: {secrets_file_path}")
            else:
                print_warning(f"Skipping: {secrets_file_path}")
        return None

    ######################################################
    ## Docker Resources
    ######################################################

    def get_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} DockerResourceGroup")

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return None
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)

        # Container Environment
        container_env: Dict[str, str] = {
            # Env variables used by data workfloapp and data assets
            "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            PHIDATA_RUNTIME_ENV_VAR: "superset",
        }

        if self.args.superset_config_path is not None:
            container_env["SUPERSET_CONFIG_PATH"] = self.args.superset_config_path

        if self.args.requirements_local is not None:
            container_env["REQUIREMENTS_LOCAL"] = self.args.requirements_local

        if self.args.python_path is not None:
            container_env["PYTHONPATH"] = self.args.python_path

        # Superset db connection
        db_user = self.args.db_user
        db_password = self.args.db_password
        db_schema = self.args.db_schema
        db_host = self.args.db_host
        db_port = self.args.db_port
        db_dialect = self.args.db_dialect
        if self.args.db_app is not None and isinstance(self.args.db_app, DbApp):
            logger.debug(f"Reading db connection details from: {self.args.db_app.name}")
            if db_user is None:
                db_user = self.args.db_app.get_db_user()
            if db_password is None:
                db_password = self.args.db_app.get_db_password()
            if db_schema is None:
                db_schema = self.args.db_app.get_db_schema()
            if db_host is None:
                db_host = self.args.db_app.get_db_host_docker()
            if db_port is None:
                db_port = self.args.db_app.get_db_port_docker()
            if db_dialect is None:
                db_dialect = self.args.db_app.get_db_driver()
        db_connection_url = (
            f"{db_dialect}://{db_user}:{db_password}@{db_host}:{db_port}/{db_schema}"
        )

        if db_user is not None:
            container_env["DATABASE_USER"] = db_user

        if db_password is not None:
            container_env["DATABASE_PASSWORD"] = db_password

        if db_schema is not None:
            container_env["DATABASE_DB"] = db_schema

        if db_host is not None:
            container_env["DATABASE_HOST"] = db_host

        if db_port is not None:
            container_env["DATABASE_PORT"] = str(db_port)

        if db_dialect is not None:
            container_env["DATABASE_DIALECT"] = db_dialect

        if self.args.flask_env is not None:
            container_env["FLASK_ENV"] = self.args.flask_env

        if self.args.superset_env is not None:
            container_env["SUPERSET_ENV"] = self.args.superset_env

        if self.args.superset_load_examples is not None:
            container_env["SUPERSET_LOAD_EXAMPLES"] = self.args.superset_load_examples

        # Superset redis connection
        redis_host = self.args.redis_host
        redis_port = self.args.redis_port
        if self.args.redis_app is not None and isinstance(self.args.redis_app, DbApp):
            logger.debug(
                f"Reading redis connection details from: {self.args.redis_app.name}"
            )
            if redis_host is None:
                redis_host = self.args.redis_app.get_db_host_docker()
            if redis_port is None:
                redis_port = self.args.redis_app.get_db_port_docker()

        if redis_host is not None:
            container_env["REDIS_HOST"] = redis_host

        if redis_port is not None:
            container_env["REDIS_PORT"] = str(redis_port)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env.update(secret_data_from_file)

        # Update the container env with user provided env, this overwrites any existing variables
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Container Volumes
        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes = self.args.container_volumes or {}
        # Create a volume for the workspace dir
        if self.args.mount_workspace:
            workspace_root_path_str = str(self.workspace_root_path)
            workspace_root_container_path_str = str(workspace_root_container_path)
            logger.debug(f"Mounting: {workspace_root_path_str}")
            logger.debug(f"\tto: {workspace_root_container_path_str}")
            container_volumes[workspace_root_path_str] = {
                "bind": workspace_root_container_path_str,
                "mode": "rw",
            }
        # Create a volume for aapp config
        if self.args.mount_resources:
            resources_dir_path = str(
                self.workspace_root_path.joinpath(self.args.resources_dir)
            )
            logger.debug(f"Mounting: {resources_dir_path}")
            logger.debug(f"\tto: {self.args.resources_dir_container_path}")
            container_volumes[resources_dir_path] = {
                "bind": self.args.resources_dir_container_path,
                "mode": "ro",
            }

        # Container Ports
        # container_ports is a dictionary which configures the ports to bind
        # inside the container. The key is the port to bind inside the container
        #   either as an integer or a string in the form port/protocol
        # and the value is the corresponding port to open on the host.
        # For example:
        #   {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
        container_ports: Dict[str, int] = {}

        # if open_container_port = True
        if self.args.open_container_port:
            container_ports[
                str(self.args.container_port)
            ] = self.args.container_host_port

        # if open_app_port = True
        # 1. Set the app_port in the container env
        # 2. Open the app_port
        if self.args.open_app_port:
            # Set the app port in the container_env
            container_env["SUPERSET_PORT"] = str(self.args.app_port)
            # Open the port
            container_ports[str(self.args.app_port)] = self.args.app_host_port

        # Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=get_image_str(self.args.image_name, self.args.image_tag),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            remove=self.args.container_remove,
            stdin_open=True,
            tty=True,
            labels=self.args.container_labels,
            environment=container_env,
            network=docker_build_context.network,
            ports=container_ports if len(container_ports) > 0 else None,
            volumes=container_volumes,
            use_cache=self.args.use_cache,
        )
        logger.debug(f"Container Env: {docker_container.environment}")

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        # Define K8s resources
        config_maps: List[CreateConfigMap] = []
        secrets: List[CreateSecret] = []
        volumes: List[CreateVolume] = []
        containers: List[CreateContainer] = []
        services: List[CreateService] = []
        ports: List[CreatePort] = []

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return None
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        requirements_file_container_path = workspace_root_container_path.joinpath(
            self.args.requirements_file_path
        )
        scripts_dir_container_path = (
            workspace_root_container_path.joinpath(self.scripts_dir)
            if self.scripts_dir
            else None
        )
        storage_dir_container_path = (
            workspace_root_container_path.joinpath(self.storage_dir)
            if self.storage_dir
            else None
        )
        meta_dir_container_path = (
            workspace_root_container_path.joinpath(self.meta_dir)
            if self.meta_dir
            else None
        )
        products_dir_container_path = (
            workspace_root_container_path.joinpath(self.products_dir)
            if self.products_dir
            else None
        )
        notebooks_dir_container_path = (
            workspace_root_container_path.joinpath(self.notebooks_dir)
            if self.notebooks_dir
            else None
        )
        workspace_config_dir_container_path = (
            workspace_root_container_path.joinpath(self.workspace_config_dir)
            if self.workspace_config_dir
            else None
        )

        # Superset db connection
        db_user = self.args.db_user
        db_password = self.args.db_password
        db_schema = self.args.db_schema
        db_host = self.args.db_host
        db_port = self.args.db_port
        db_driver = self.args.db_driver
        if self.args.db_app is not None and isinstance(self.args.db_app, DbApp):
            logger.debug(f"Reading db connection details from: {self.args.db_app.name}")
            if db_user is None:
                db_user = self.args.db_app.get_db_user()
            if db_password is None:
                db_password = self.args.db_app.get_db_password()
            if db_schema is None:
                db_schema = self.args.db_app.get_db_schema()
            if db_host is None:
                db_host = self.args.db_app.get_db_host_k8s()
            if db_port is None:
                db_port = self.args.db_app.get_db_port_k8s()
            if db_driver is None:
                db_driver = self.args.db_app.get_db_driver()
        db_connection_url = (
            f"{db_driver}://{db_user}:{db_password}@{db_host}:{db_port}/{db_schema}"
        )

        # Container pythonpath
        python_path = self.args.python_path or str(workspace_root_container_path)

        # Container Environment
        container_env: Dict[str, str] = {
            # Env variables used by data workfloapp and data assets
            "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            "PYTHONPATH": python_path,
            PHIDATA_RUNTIME_ENV_VAR: "superset",
            SCRIPTS_DIR_ENV_VAR: str(scripts_dir_container_path),
            STORAGE_DIR_ENV_VAR: str(storage_dir_container_path),
            META_DIR_ENV_VAR: str(meta_dir_container_path),
            PRODUCTS_DIR_ENV_VAR: str(products_dir_container_path),
            NOTEBOOKS_DIR_ENV_VAR: str(notebooks_dir_container_path),
            WORKSPACE_CONFIG_DIR_ENV_VAR: str(workspace_config_dir_container_path),
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": str(requirements_file_container_path),
            "MOUNT_WORKSPACE": str(self.args.mount_workspace),
            # Print env when the container starts
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            # Env variables used by Superset
            # INIT_AIRFLOW env var is required for phidata to create DAGs
            "INIT_AIRFLOW": str(True),
            "AIRFLOW_ENV": self.args.superset_env,
            "WAIT_FOR_DB": str(self.args.wait_for_db),
            "WAIT_FOR_DB_INIT": str(self.args.wait_for_db_init),
            "INIT_AIRFLOW_DB": str(self.args.init_superset_db),
            "UPGRADE_AIRFLOW_DB": str(self.args.upgrade_superset_db),
            "DB_USER": str(db_user),
            "DB_PASSWORD": str(db_password),
            "DB_SCHEMA": str(db_schema),
            "DB_HOST": str(db_host),
            "DB_PORT": str(db_port),
            "WAIT_FOR_REDIS": str(self.args.wait_for_redis),
            "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            "CREATE_AIRFLOW_ADMIN_USER": str(self.args.create_superset_admin_user),
            "AIRFLOW__CORE__EXECUTOR": str(self.args.executor),
        }

        # Set the AIRFLOW__DATABASE__SQL_ALCHEMY_CONN
        if "None" not in db_connection_url:
            logger.debug(f"AIRFLOW__DATABASE__SQL_ALCHEMY_CONN: {db_connection_url}")
            container_env["AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"] = db_connection_url

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_products_as_superset_dags:
            container_env["AIRFLOW__CORE__DAGS_FOLDER"] = str(
                products_dir_container_path
            )
        elif self.args.superset_dags_path is not None:
            container_env["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.superset_dags_path

        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    container_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue

        if self.args.executor == "CeleryExecutor":
            # Superset celery result backend
            celery_result_backend_driver = (
                self.args.db_result_backend_driver or db_driver
            )
            celery_result_backend_url = f"{celery_result_backend_driver}://{db_user}:{db_password}@{db_host}:{db_port}/{db_schema}"
            # Set the AIRFLOW__CELERY__RESULT_BACKEND
            if "None" not in celery_result_backend_url:
                container_env[
                    "AIRFLOW__CELERY__RESULT_BACKEND"
                ] = celery_result_backend_url

            # Superset celery broker url
            redis_password = (
                f"{self.args.redis_password}@" if self.args.redis_password else ""
            )
            redis_schema = self.args.redis_schema
            redis_host = self.args.redis_host
            redis_port = self.args.redis_port
            redis_driver = self.args.redis_driver
            if self.args.redis_app is not None and isinstance(
                self.args.redis_app, DbApp
            ):
                logger.debug(
                    f"Reading redis connection details from: {self.args.redis_app.name}"
                )
                if redis_password is None:
                    redis_password = self.args.redis_app.get_db_password()
                if redis_schema is None:
                    redis_schema = self.args.redis_app.get_db_schema() or "0"
                if redis_host is None:
                    redis_host = self.args.redis_app.get_db_host_k8s()
                if redis_port is None:
                    redis_port = self.args.redis_app.get_db_port_k8s()
                if redis_driver is None:
                    redis_driver = self.args.redis_app.get_db_driver()

            # Set the AIRFLOW__CELERY__RESULT_BACKEND
            celery_broker_url = f"{redis_driver}://{redis_password}{redis_host}:{redis_port}/{redis_schema}"
            if "None" not in celery_broker_url:
                container_env["AIRFLOW__CELERY__BROKER_URL"] = celery_broker_url

            # Set the redis connection details
            if redis_password is not None:
                container_env["REDIS_PASSWORD"] = redis_password
            if redis_schema is not None:
                container_env["REDIS_SCHEMA"] = redis_schema
            if redis_host is not None:
                container_env["REDIS_HOST"] = redis_host
            if redis_port is not None:
                container_env["REDIS_PORT"] = str(redis_port)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.config_map_name or get_default_configmap_name(app_name),
            app_name=app_name,
            data=container_env,
        )
        # logger.debug(f"ConfigMap {container_env_cm.cm_name}: {container_env_cm.json(indent=2)}")
        config_maps.append(container_env_cm)

        # Create a Secret to set the container env variables which are Secret
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env_secret = CreateSecret(
                secret_name=self.args.secret_name or get_default_secret_name(app_name),
                app_name=app_name,
                string_data=secret_data_from_file,
            )
            secrets.append(container_env_secret)

        # If mount_workspace=True first check if the workspace
        # should be mounted locally, otherwise
        # Create a Sidecar git-sync container and volume
        if self.args.mount_workspace:
            workspace_volume_name = (
                self.args.workspace_volume_name or get_default_volume_name(app_name)
            )

            if self.args.k8s_mount_local_workspace:
                workspace_root_path_str = str(self.workspace_root_path)
                workspace_root_container_path_str = str(workspace_root_container_path)
                logger.debug(f"Mounting: {workspace_root_path_str}")
                logger.debug(f"\tto: {workspace_root_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_root_container_path_str,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=workspace_root_path_str,
                    ),
                )
                volumes.append(workspace_volume)

            elif self.args.create_git_sync_sidecar:
                workspace_parent_container_path_str = str(
                    self.args.workspace_parent_container_path
                )
                logger.debug(f"Creating EmptyDir")
                logger.debug(f"\tat: {workspace_parent_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_parent_container_path_str,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                volumes.append(workspace_volume)

                if self.args.git_sync_repo is None:
                    print_error("git_sync_repo invalid")
                else:
                    git_sync_env = {
                        "GIT_SYNC_REPO": self.args.git_sync_repo,
                        "GIT_SYNC_ROOT": str(self.args.workspace_parent_container_path),
                        "GIT_SYNC_DEST": workspace_name,
                    }
                    if self.args.git_sync_branch is not None:
                        git_sync_env["GIT_SYNC_BRANCH"] = self.args.git_sync_branch
                    if self.args.git_sync_wait is not None:
                        git_sync_env["GIT_SYNC_WAIT"] = str(self.args.git_sync_wait)
                    git_sync_sidecar = CreateContainer(
                        container_name="git-sync-workspaces",
                        app_name=app_name,
                        image_name="k8s.gcr.io/git-sync",
                        image_tag="v3.1.1",
                        env=git_sync_env,
                        envs_from_configmap=[cm.cm_name for cm in config_maps]
                        if len(config_maps) > 0
                        else None,
                        envs_from_secret=[secret.secret_name for secret in secrets]
                        if len(secrets) > 0
                        else None,
                        volumes=[workspace_volume],
                    )
                    containers.append(git_sync_sidecar)

        # Create the ports to open
        # if open_container_port = True
        if self.args.open_container_port:
            container_port = CreatePort(
                name=self.args.container_port_name,
                container_port=self.args.container_port,
            )
            ports.append(container_port)

        # if open_app_port = True
        # 1. Set the app_port in the container env
        # 2. Open the superset app port
        app_port: Optional[CreatePort] = None
        if self.args.open_app_port:
            # Set the app port in the container env
            if container_env_cm.data is not None:
                container_env_cm.data["AIRFLOW__WEBSERVER__WEB_SERVER_PORT"] = str(
                    self.args.app_port
                )
            # Open the port
            app_port = CreatePort(
                name=self.args.app_port_name,
                container_port=self.args.app_port,
                service_port=self.get_app_service_port(),
                node_port=self.args.app_node_port,
                target_port=self.args.app_target_port or self.args.app_port_name,
            )
            ports.append(app_port)

        # if open_worker_log_port = True
        # 1. Set the worker_log_port in the container env
        # 2. Open the worker_log_port
        if self.args.open_worker_log_port:
            # Set the worker_log_port in the container_env
            if container_env_cm.data is not None:
                container_env_cm.data["AIRFLOW__LOGGING__WORKER_LOG_SERVER_PORT"] = str(
                    self.args.worker_log_port
                )
            # Open the port
            worker_log_port = CreatePort(
                name=self.args.worker_log_port_name,
                container_port=self.args.worker_log_port,
            )
            ports.append(worker_log_port)

        # if open_flower_port = True
        # 1. Set the flower_port in the container env
        # 2. Open the flower_port
        flower_port: Optional[CreatePort] = None
        if self.args.open_flower_port:
            # Set the flower_port in the container_env
            if container_env_cm.data is not None:
                container_env_cm.data["AIRFLOW__CELERY__FLOWER_PORT"] = str(
                    self.args.flower_port
                )
            # Open the port
            flower_port = CreatePort(
                name=self.args.flower_port_name,
                container_port=self.args.flower_port,
                service_port=self.get_flower_service_port(),
                target_port=self.args.flower_target_port or self.args.flower_port_name,
            )
            ports.append(flower_port)

        container_labels: Optional[Dict[str, Any]] = self.args.container_labels
        if k8s_build_context.labels is not None:
            if container_labels:
                container_labels.update(k8s_build_context.labels)
            else:
                container_labels = k8s_build_context.labels
        # Create the container
        k8s_container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=[self.args.command]
            if isinstance(self.args.command, str)
            else self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            image_pull_policy=self.args.image_pull_policy,
            envs_from_configmap=[cm.cm_name for cm in config_maps]
            if len(config_maps) > 0
            else None,
            envs_from_secret=[secret.secret_name for secret in secrets]
            if len(secrets) > 0
            else None,
            ports=ports if len(ports) > 0 else None,
            volumes=volumes if len(volumes) > 0 else None,
            labels=container_labels,
        )
        containers.append(k8s_container)

        # Set default container for kubectl commands
        # https://kubernetes.io/docs/reference/labels-annotations-taints/#kubectl-kubernetes-io-default-container
        pod_annotations = {
            "kubectl.kubernetes.io/default-container": k8s_container.container_name
        }

        deploy_labels: Optional[Dict[str, Any]] = self.args.deploy_labels
        if k8s_build_context.labels is not None:
            if deploy_labels:
                deploy_labels.update(k8s_build_context.labels)
            else:
                deploy_labels = k8s_build_context.labels
        # Create the deployment
        k8s_deployment = CreateDeployment(
            replicas=self.args.replicas,
            deploy_name=self.args.deploy_name or get_default_deploy_name(app_name),
            pod_name=self.args.pod_name or get_default_pod_name(app_name),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            containers=containers if len(containers) > 0 else None,
            pod_node_selector=self.args.pod_node_selector,
            restart_policy=self.args.restart_policy,
            termination_grace_period_seconds=self.args.termination_grace_period_seconds,
            volumes=volumes if len(volumes) > 0 else None,
            labels=deploy_labels,
            pod_annotations=pod_annotations,
            topology_spread_key=self.args.topology_spread_key,
            topology_spread_max_skew=self.args.topology_spread_max_skew,
            topology_spread_when_unsatisfiable=self.args.topology_spread_when_unsatisfiable,
        )

        # Create the services
        if self.args.create_app_service:
            app_service_labels: Optional[Dict[str, Any]] = self.args.app_service_labels
            if k8s_build_context.labels is not None:
                if app_service_labels:
                    app_service_labels.update(k8s_build_context.labels)
                else:
                    app_service_labels = k8s_build_context.labels
            app_service = CreateService(
                service_name=self.get_app_service_name(),
                app_name=app_name,
                namespace=k8s_build_context.namespace,
                service_account_name=k8s_build_context.service_account_name,
                service_type=self.args.app_service_type,
                deployment=k8s_deployment,
                ports=[app_port] if app_port else None,
                labels=app_service_labels,
            )
            services.append(app_service)

        if self.args.create_flower_service:
            flower_service_labels: Optional[
                Dict[str, Any]
            ] = self.args.flower_service_labels
            if k8s_build_context.labels is not None:
                if flower_service_labels:
                    flower_service_labels.update(k8s_build_context.labels)
                else:
                    flower_service_labels = k8s_build_context.labels
            flower_service = CreateService(
                service_name=self.get_flower_service_name(),
                app_name=app_name,
                namespace=k8s_build_context.namespace,
                service_account_name=k8s_build_context.service_account_name,
                service_type=self.args.flower_service_type,
                deployment=k8s_deployment,
                ports=[flower_port] if flower_port else None,
                labels=flower_service_labels,
            )
            services.append(flower_service)

        # Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            config_maps=config_maps if len(config_maps) > 0 else None,
            secrets=secrets if len(secrets) > 0 else None,
            services=services if len(services) > 0 else None,
            deployments=[k8s_deployment],
        )

        return k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg = self.get_k8s_rg(k8s_build_context)
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



# Need to be sure that we dont import anything which may lead to circluar imports
from typing import Any, List, Optional, Type


def get_default_ns_name(app_name: str) -> str:
    return "{}-ns".format(app_name)


def get_default_ctx_name(app_name: str) -> str:
    return "{}-ctx".format(app_name)


def get_default_sa_name(app_name: str) -> str:
    return "{}-sa".format(app_name)


def get_default_cr_name(app_name: str) -> str:
    return "{}-cr".format(app_name)


def get_default_crb_name(app_name: str) -> str:
    return "{}-crb".format(app_name)


def get_default_pod_name(app_name: str) -> str:
    return "{}-pod".format(app_name)


def get_default_container_name(app_name: str) -> str:
    return "{}-container".format(app_name)


def get_default_service_name(app_name: str) -> str:
    return "{}-svc".format(app_name)


def get_default_deploy_name(app_name: str) -> str:
    return "{}-deploy".format(app_name)


def get_default_configmap_name(app_name: str) -> str:
    return "{}-cm".format(app_name)


def get_default_secret_name(app_name: str) -> str:
    return "{}-secret".format(app_name)


def get_default_volume_name(app_name: str) -> str:
    return "{}-volume".format(app_name)


def get_default_pvc_name(app_name: str) -> str:
    return "{}-pvc".format(app_name)


def get_image_str(repo: str, tag: str) -> str:
    return f"{repo}:{tag}"


def isinstanceany(obj: Any, class_list: List[Type]) -> bool:
    for cls in class_list:
        if isinstance(obj, cls):
            return True
    return False


def str_to_int(inp: str) -> Optional[int]:
    """
    Safely converts a string value to integer.
    Args:
        inp: input string

    Returns: input string as int if possible, None if not
    """
    try:
        val = int(inp)
        return val
    except Exception:
        return None


def is_empty(val: Any) -> bool:
    if val is None or len(val) == 0 or val == "":
        return True
    return False



from typing import Dict, Optional


def create_component_labels_dict(
    component_name: str, app_name: str, labels: Optional[Dict[str, str]] = None
) -> Dict[str, str]:
    _labels = {
        "app.kubernetes.io/component": component_name,
        "app.kubernetes.io/app": app_name,
    }
    if labels:
        _labels.update(labels)

    return _labels



from typing import Optional
from functools import wraps, partial

from phidata.task.python_task import PythonTask, PythonTaskType, PythonTaskDecoratorType
from phidata.utils.log import logger


def task(func: PythonTaskType) -> PythonTaskDecoratorType:
    """Converts a function into a PythonTask"""

    wraps(func)

    def decorator(
        *args,
        name: Optional[str] = None,
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> PythonTask:
        # logger.info(f"task func: {func}")
        # logger.info(f"task kwargs: {kwargs}")
        task_name = name or func.__name__
        try:
            # create a partial function that can be called by the PythonTask
            entrypoint = partial(func, *args, **kwargs)
            python_task = PythonTask(
                name=task_name,
                entrypoint=entrypoint,
                task_id=task_id,
                dag_id=dag_id,
                entrypoint_args=args,
                entrypoint_kwargs=kwargs,
                version=version,
                enabled=enabled,
            )
            return python_task
        except Exception as e:
            logger.error(f"Could not create task: {task_name}")
            logger.error(e)
            raise

    return decorator



from typing import Optional, Any, List, Dict

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.iam.policy import IamPolicy
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class IamRole(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/iam.html#service-resource
    """

    resource_type = "IamRole"
    service_name = "iam"

    # RoleName: The name of the role to create.
    name: str
    # The trust relationship policy document that grants an entity permission to assume the role.
    assume_role_policy_document: str
    # The path to the role. This parameter is optional. If it is not included, it defaults to a slash (/).
    path: Optional[str] = None
    # A description of the role.
    description: Optional[str] = None
    # The maximum session duration (in seconds) that you want to set for the specified role.
    # If you do not specify a value for this setting, the default maximum of one hour is applied.
    # This setting can have a value from 1 hour to 12 hours.
    max_session_duration: Optional[int] = None
    # The ARN of the policy that is used to set the permissions boundary for the role.
    permissions_boundary: Optional[str] = None
    # A list of tags that you want to attach to the new role. Each tag consists of a key name and an associated value.
    tags: Optional[List[Dict[str, str]]] = None

    # List of IAM policies to
    # attach to the role after it is created
    policies: Optional[List[IamPolicy]] = None
    # List of IAM policy ARNs (Amazon Resource Name) to
    # attach to the role after it is created
    policy_arns: Optional[List[str]] = None

    # The Amazon Resource Name (ARN) specifying the role.
    # To get the arn, use get_arn() function
    arn: Optional[str] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the IamRole

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args: Dict[str, Any] = {}
            if self.path:
                not_null_args["Path"] = self.path
            if self.description:
                not_null_args["Description"] = self.description
            if self.max_session_duration:
                not_null_args["MaxSessionDuration"] = self.max_session_duration
            if self.permissions_boundary:
                not_null_args["PermissionsBoundary"] = self.permissions_boundary
            if self.tags:
                not_null_args["Tags"] = self.tags

            # Create Role
            service_resource = self.get_service_resource(aws_client)
            role = service_resource.create_role(
                RoleName=self.name,
                AssumeRolePolicyDocument=self.assume_role_policy_document,
                **not_null_args,
            )
            # logger.debug(f"Role: {role}")

            # Validate Role creation
            create_date = role.create_date
            self.arn = role.arn
            logger.debug(f"create_date: {create_date}")
            logger.debug(f"arn: {self.arn}")
            if create_date is not None:
                print_info(f"Role created: {self.name}")
                self.active_resource = role
                return True
            print_error("Role could not be created")
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for Role to be created
        if self.wait_for_creation:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be created.")
                waiter = self.get_service_client(aws_client).get_waiter("role_exists")
                waiter.wait(
                    RoleName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error("Waiter failed.")
                print_error(e)
        # Attach policy arns to role
        if self.active_resource is not None and self.policy_arns is not None:
            self.attach_policy_arns(aws_client)
        # Attach policies to role
        if self.active_resource is not None and self.policies is not None:
            self.attach_policies(aws_client)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the IamRole

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_resource = self.get_service_resource(aws_client)
            role = service_resource.Role(name=self.name)
            role.load()
            create_date = role.create_date
            self.arn = role.arn
            logger.debug(f"create_date: {create_date}")
            logger.debug(f"arn: {self.arn}")
            if create_date is not None:
                logger.debug(f"Role found: {role.role_name}")
                self.active_resource = role
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
            pass
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the IamRole

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            role = self._read(aws_client)
            # logger.debug(f"Role: {role}")
            # logger.debug(f"Role type: {type(role)}")
            self.active_resource = None

            if role is None:
                logger.warning(f"No {self.get_resource_type()} to delete")
                return True

            # detach all policies
            policies = role.attached_policies.all()
            for policy in policies:
                print_info(f"Detaching policy: {policy}")
                role.detach_policy(PolicyArn=policy.arn)

            # detach all instance profiles
            profiles = role.instance_profiles.all()
            for profile in profiles:
                print_info(f"Removing role from profile: {profile}")
                profile.remove_role(RoleName=role.name)

            # delete role
            role.delete()
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def attach_policy_arns(self, aws_client: AwsApiClient) -> bool:
        """
        Attaches the specified managed policy to the specified IAM role.
        When you attach a managed policy to a role, the managed policy becomes part of the
        role's permission (access) policy.

        Returns:
            True if operation was successful
        """
        if self.policy_arns is None:
            return True

        role = self._read(aws_client)
        if role is None:
            logger.warning(f"No {self.get_resource_type()} to attach")
            return True
        try:
            # logger.debug("Attaching managed policies to role")
            for arn in self.policy_arns:
                if isinstance(arn, str):
                    role.attach_policy(PolicyArn=arn)
                    print_info(f"Attaching policy to {role.role_name}: {arn}")
            return True
        except Exception as e:
            print_error(e)
        return False

    def attach_policies(self, aws_client: AwsApiClient) -> bool:
        """
        Returns:
            True if operation was successful
        """
        if self.policies is None:
            return True

        role = self._read(aws_client)
        if role is None:
            logger.warning(f"No {self.get_resource_type()} to attach")
            return True
        try:
            logger.debug("Attaching managed policies to role")
            for policy in self.policies:
                if policy.arn is None:
                    policy.create(aws_client)
                if policy.arn is not None:
                    role.attach_policy(PolicyArn=policy.arn)
                    print_info(f"Attaching policy to {role.role_name}: {policy.arn}")
            return True
        except Exception as e:
            print_error(e)
        return False

    def get_arn(self, aws_client: AwsApiClient) -> Optional[str]:
        role = self._read(aws_client)
        if role is None:
            return None

        self.arn = role.arn
        return self.arn



from typing import List, Optional

from kubernetes.client import RbacAuthorizationV1Api
from kubernetes.client.models.v1_cluster_role_binding import V1ClusterRoleBinding
from kubernetes.client.models.v1_cluster_role_binding_list import (
    V1ClusterRoleBindingList,
)
from kubernetes.client.models.v1_role_ref import V1RoleRef
from kubernetes.client.models.v1_subject import V1Subject
from kubernetes.client.models.v1_status import V1Status
from pydantic import Field

from phidata.infra.k8s.enums.api_group import ApiGroup
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource, K8sObject
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class Subject(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#subject-v1-rbac-authorization-k8s-io
    """

    resource_type: str = "Subject"

    # Name of the object being referenced.
    name: str
    # Kind of object being referenced.
    # Values defined by this API group are "User", "Group", and "ServiceAccount".
    # If the Authorizer does not recognized the kind value, the Authorizer should report an error.
    kind: Kind
    # Namespace of the referenced object.
    # If the object kind is non-namespace, such as "User" or "Group", and this value is not empty
    # the Authorizer should report an error.
    namespace: str
    # APIGroup holds the API group of the referenced subject.
    # Defaults to "" for ServiceAccount subjects.
    # Defaults to "rbac.authorization.k8s.io" for User and Group subjects.
    api_group: Optional[ApiGroup] = Field(None, alias="apiGroup")

    def get_k8s_object(self) -> V1Subject:
        # Return a V1Subject object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_subject.py
        _v1_subject = V1Subject(
            api_group=self.api_group,
            kind=self.kind,
            name=self.name,
            namespace=self.namespace,
        )
        return _v1_subject


class RoleRef(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#roleref-v1-rbac-authorization-k8s-io
    """

    resource_type: str = "RoleRef"

    # APIGroup is the group for the resource being referenced
    api_group: ApiGroup = Field(..., alias="apiGroup")
    # Kind is the type of resource being referenced
    kind: Kind
    # Name is the name of resource being referenced
    name: str

    def get_k8s_object(self) -> V1RoleRef:
        # Return a V1RoleRef object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_role_ref.py
        _v1_role_ref = V1RoleRef(
            api_group=self.api_group,
            kind=self.kind,
            name=self.name,
        )
        return _v1_role_ref


class ClusterRoleBinding(K8sResource):
    """
    References:
        * Doc: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#clusterrolebinding-v1-rbac-authorization-k8s-io
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_cluster_role_binding_binding.py
    """

    resource_type: str = "ClusterRoleBinding"

    role_ref: RoleRef = Field(..., alias="roleRef")
    subjects: List[Subject]

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["roleRef", "subjects"]

    # V1ClusterRoleBinding object received as the output after creating the crb
    v1_cluster_role_binding: Optional[V1ClusterRoleBinding] = None

    def get_k8s_object(self) -> V1ClusterRoleBinding:
        """Creates a body for this ClusterRoleBinding"""

        subjects_list = None
        if self.subjects:
            subjects_list = []
            for subject in self.subjects:
                subjects_list.append(subject.get_k8s_object())

        # Return a V1ClusterRoleBinding object to create a ClusterRoleBinding
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_cluster_role_binding.py
        _v1_cluster_role_binding = V1ClusterRoleBinding(
            api_version=self.api_version,
            kind=self.kind,
            metadata=self.metadata.get_k8s_object(),
            role_ref=self.role_ref.get_k8s_object(),
            subjects=subjects_list,
        )
        return _v1_cluster_role_binding

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1ClusterRoleBinding]]:
        """Reads ClusterRoles from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: NOT USED.
        """
        rbac_auth_v1_api: RbacAuthorizationV1Api = k8s_client.rbac_auth_v1_api
        crb_list: Optional[
            V1ClusterRoleBindingList
        ] = rbac_auth_v1_api.list_cluster_role_binding()
        crbs: Optional[List[V1ClusterRoleBinding]] = None
        if crb_list:
            crbs = crb_list.items
            # logger.debug(f"crbs: {crbs}")
            # logger.debug(f"crbs type: {type(crbs)}")
        return crbs

    def _create(self, k8s_client: K8sApiClient) -> bool:

        rbac_auth_v1_api: RbacAuthorizationV1Api = k8s_client.rbac_auth_v1_api
        k8s_object: V1ClusterRoleBinding = self.get_k8s_object()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_cluster_role_binding: V1ClusterRoleBinding = (
            rbac_auth_v1_api.create_cluster_role_binding(
                body=k8s_object,
                async_req=self.async_req,
                pretty=self.pretty,
            )
        )
        # logger.debug("Created: {}".format(v1_cluster_role_binding))
        if v1_cluster_role_binding.metadata.creation_timestamp is not None:
            logger.debug("ClusterRoleBinding Created")
            self.active_resource = v1_cluster_role_binding
            self.active_resource_class = V1ClusterRoleBinding
            return True
        logger.error("ClusterRoleBinding could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1ClusterRoleBinding]:
        """Returns the "Active" ClusterRoleBinding from the cluster"""

        active_resource: Optional[V1ClusterRoleBinding] = None
        active_resources: Optional[List[V1ClusterRoleBinding]] = self.get_from_cluster(
            k8s_client=k8s_client,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_crb.metadata.name: _crb for _crb in active_resources}

        crb_name = self.get_resource_name()
        if crb_name in active_resources_dict:
            active_resource = active_resources_dict[crb_name]
            self.active_resource = active_resource
            self.active_resource_class = V1ClusterRoleBinding
            logger.debug(f"Found active {crb_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:

        rbac_auth_v1_api: RbacAuthorizationV1Api = k8s_client.rbac_auth_v1_api
        crb_name = self.get_resource_name()
        k8s_object: V1ClusterRoleBinding = self.get_k8s_object()

        logger.debug("Updating: {}".format(crb_name))
        v1_cluster_role_binding: V1ClusterRoleBinding = (
            rbac_auth_v1_api.patch_cluster_role_binding(
                name=crb_name,
                body=k8s_object,
                async_req=self.async_req,
                pretty=self.pretty,
            )
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_cluster_role_binding.to_dict(), indent=2)))
        if v1_cluster_role_binding.metadata.creation_timestamp is not None:
            logger.debug("ClusterRoleBinding Updated")
            self.active_resource = v1_cluster_role_binding
            self.active_resource_class = V1ClusterRoleBinding
            return True
        logger.error("ClusterRoleBinding could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        rbac_auth_v1_api: RbacAuthorizationV1Api = k8s_client.rbac_auth_v1_api
        crb_name = self.get_resource_name()

        logger.debug("Deleting: {}".format(crb_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = rbac_auth_v1_api.delete_cluster_role_binding(
            name=crb_name,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("delete_status: {}".format(delete_status.status))
        if delete_status.status == "Success":
            logger.debug("ClusterRoleBinding Deleted")
            return True
        logger.error("ClusterRoleBinding could not be deleted")
        return False



import json
from typing import List, Optional, Dict, Any

from phidata.infra.aws.resource.iam.role import IamRole, IamPolicy
from phidata.infra.aws.resource.s3.bucket import S3Bucket


def create_glue_iam_policy(
    name: str = "glueS3CrawlerPolicy",
    version: str = "2022-04-26",
    s3_buckets: Optional[List[S3Bucket]] = None,
) -> IamPolicy:
    statements: List[Dict[str, Any]] = []
    if s3_buckets is not None:
        statements.append(
            {
                "Effect": "Allow",
                "Action": ["s3:ListBucket"],
                "Resource": [f"arn:aws:s3:::{bucket.name}" for bucket in s3_buckets],
            }
        )
        statements.append(
            {
                "Effect": "Allow",
                "Action": ["s3:GetObject"],
                "Resource": [f"arn:aws:s3:::{bucket.name}/*" for bucket in s3_buckets],
            }
        )
    else:
        statements.append(
            {
                "Effect": "Allow",
                "Action": ["s3:ListBucket"],
                "Resource": ["arn:aws:s3:::*"],
            }
        )
        statements.append(
            {
                "Effect": "Allow",
                "Action": ["s3:GetObject"],
                "Resource": ["arn:aws:s3:::*/*"],
            }
        )

    return IamPolicy(
        name=name,
        policy_document=json.dumps(
            {
                "Version": version,
                "Statement": statements,
            }
        ),
    )


def create_glue_iam_role(
    name: str,
    s3_buckets: Optional[List[S3Bucket]] = None,
    iam_policy_name: str = "glueS3CrawlerPolicy",
    iam_policy_version: str = "2022-04-26",
    skip_create: bool = False,
    skip_delete: bool = False,
) -> IamRole:

    iam_policy = create_glue_iam_policy(
        name=iam_policy_name, version=iam_policy_version, s3_buckets=s3_buckets
    )
    return IamRole(
        name=name,
        assume_role_policy_document=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "glue.amazonaws.com"},
                        "Action": "sts:AssumeRole",
                    }
                ],
            }
        ),
        policies=[iam_policy],
        policy_arns=["arn:aws:iam::aws:policy/service-role/AWSGlueServiceRole"],
        skip_create=skip_create,
        skip_delete=skip_delete,
    )



from typing import List, Optional, Any, Dict
from typing_extensions import Literal

from kubernetes.client import ApiextensionsV1Api
from kubernetes.client.models.v1_custom_resource_definition import (
    V1CustomResourceDefinition,
)
from kubernetes.client.models.v1_custom_resource_definition_list import (
    V1CustomResourceDefinitionList,
)
from kubernetes.client.models.v1_custom_resource_definition_names import (
    V1CustomResourceDefinitionNames,
)
from kubernetes.client.models.v1_custom_resource_definition_spec import (
    V1CustomResourceDefinitionSpec,
)
from kubernetes.client.models.v1_custom_resource_definition_version import (
    V1CustomResourceDefinitionVersion,
)
from kubernetes.client.models.v1_custom_resource_validation import (
    V1CustomResourceValidation,
)
from kubernetes.client.models.v1_json_schema_props import V1JSONSchemaProps
from kubernetes.client.models.v1_status import V1Status
from pydantic import Field

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource, K8sObject
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class CustomResourceDefinitionNames(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition_names.py
    """

    resource_type: str = "CustomResourceDefinitionNames"

    # categories is a list of grouped resources this custom resource belongs to (e.g. 'all').
    # This is published in API discovery documents, and used by clients to support invocations like `kubectl get all`.
    categories: Optional[List[str]] = None
    # kind is the serialized kind of the resource. It is normally CamelCase and singular.
    # Custom resource instances will use this value as the `kind` attribute in API calls.
    kind: str
    # listKind is the serialized kind of the list for this resource.
    # Defaults to "`kind`List".
    list_kind: Optional[str] = Field(None, alias="listKind")
    # plural is the plural name of the resource to serve.
    # The custom resources are served under `/apis/<group>/<version>/.../<plural>`.
    # Must match the name of the CustomResourceDefinition (in the form `<names.plural>.<group>`).
    # Must be all lowercase.
    plural: Optional[str] = None
    # shortNames are short names for the resource, exposed in API discovery documents,
    # and used by clients to support invocations like `kubectl get <shortname>`.
    # It must be all lowercase.
    short_names: Optional[List[str]] = Field(None, alias="shortNames")
    # singular is the singular name of the resource. It must be all lowercase.
    # Defaults to lowercased `kind`.
    singular: Optional[str] = None

    def get_k8s_object(
        self,
    ) -> V1CustomResourceDefinitionNames:

        # Return a V1CustomResourceDefinitionNames object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition_names.py
        _v1_custom_resource_definition_names = V1CustomResourceDefinitionNames(
            categories=self.categories,
            kind=self.kind,
            list_kind=self.list_kind,
            plural=self.plural,
            short_names=self.short_names,
            singular=self.singular,
        )
        return _v1_custom_resource_definition_names


class CustomResourceDefinitionVersion(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition_version.py
    """

    resource_type: str = "CustomResourceDefinitionVersion"

    # name is the version name, e.g. “v1”, “v2beta1”, etc.
    # The custom resources are served under this version at `/apis/<group>/<version>/...` if `served` is true.
    name: str
    # served is a flag enabling/disabling this version from being served via REST APIs
    served: bool = True
    # storage indicates this version should be used when persisting custom resources to storage.
    # There must be exactly one version with storage=true.
    storage: bool = True
    # schema describes the schema used for validation, pruning, and defaulting of this version of the custom resource.
    # openAPIV3Schema is the OpenAPI v3 schema to use for validation and pruning.
    open_apiv3_schema: Optional[V1JSONSchemaProps] = Field(
        None, alias="openAPIV3Schema"
    )
    # deprecated indicates this version of the custom resource API is deprecated. When set to true,
    # API requests to this version receive a warning header in the server response. Defaults to false.
    deprecated: Optional[bool] = None
    # deprecationWarning overrides the default warning returned to API clients.
    # May only be set when `deprecated` is true. The default warning indicates this version is deprecated
    # and recommends use of the newest served version of equal or greater stability, if one exists.
    deprecation_warning: Optional[str] = Field(None, alias="openAPIV3Schema")

    def get_k8s_object(
        self,
    ) -> V1CustomResourceDefinitionVersion:

        # Return a V1CustomResourceDefinitionVersion object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition_version.py
        _v1_custom_resource_definition_version = V1CustomResourceDefinitionVersion(
            # additional_printer_columns=self.additional_printer_columns,
            deprecated=self.deprecated,
            deprecation_warning=self.deprecation_warning,
            name=self.name,
            schema=V1CustomResourceValidation(
                open_apiv3_schema=self.open_apiv3_schema,
            ),
            served=self.served,
            storage=self.storage,
            # subresources=self.subresources,
        )
        return _v1_custom_resource_definition_version


class CustomResourceDefinitionSpec(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition_spec.py
    """

    resource_type: str = "CustomResourceDefinitionSpec"

    group: str
    names: CustomResourceDefinitionNames
    preserve_unknown_fields: Optional[bool] = Field(None, alias="preserveUnknownFields")
    # scope indicates whether the defined custom resource is cluster- or namespace-scoped.
    # Allowed values are `Cluster` and `Namespaced`.
    scope: Literal["Cluster", "Namespaced"]
    # versions is the list of all API versions of the defined custom resource.
    # Version names are used to compute the order in which served versions are listed in API discovery.
    # If the version string is "kube-like", it will sort above non "kube-like" version strings,
    # which are ordered lexicographically. "Kube-like" versions start with a "v", then are followed by a number
    # (the major version), then optionally the string "alpha" or "beta" and another number
    # (the minor version). These are sorted first by GA > beta > alpha
    # (where GA is a version with no suffix such as beta or alpha),
    # and then by comparing major version, then minor version.
    # An example sorted list of versions: v10, v2, v1, v11beta2, v10beta3, v3beta1, v12alpha1, v11alpha2, foo1, foo10.
    versions: List[CustomResourceDefinitionVersion]

    def get_k8s_object(
        self,
    ) -> V1CustomResourceDefinitionSpec:

        # Return a V1CustomResourceDefinitionSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition_spec.py
        _v1_custom_resource_definition_spec = V1CustomResourceDefinitionSpec(
            group=self.group,
            names=self.names.get_k8s_object(),
            scope=self.scope,
            versions=[version.get_k8s_object() for version in self.versions],
        )
        return _v1_custom_resource_definition_spec


class CustomResourceDefinition(K8sResource):
    """
    References:
        * Doc: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#customresourcedefinition-v1-apiextensions-k8s-io
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition.py
    """

    resource_type: str = "CustomResourceDefinition"

    spec: CustomResourceDefinitionSpec

    # List of fields to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["spec"]

    def get_k8s_object(self) -> V1CustomResourceDefinition:
        """Creates a body for this CustomResourceDefinition"""

        # Return a V1CustomResourceDefinition object to create a CustomResourceDefinition
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_custom_resource_definition.py
        _v1_custom_resource_definition = V1CustomResourceDefinition(
            api_version=self.api_version,
            kind=self.kind,
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object(),
        )
        return _v1_custom_resource_definition

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1CustomResourceDefinition]]:
        """Reads CustomResourceDefinitions from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        logger.debug(f"Getting CRDs from cluster")
        apiextensions_v1_api: ApiextensionsV1Api = k8s_client.apiextensions_v1_api
        crd_list: Optional[
            V1CustomResourceDefinitionList
        ] = apiextensions_v1_api.list_custom_resource_definition()
        crds: Optional[List[V1CustomResourceDefinition]] = None
        if crd_list:
            crds = crd_list.items
            # logger.debug(f"crds: {crds}")
            # logger.debug(f"crds type: {type(crds)}")
        return crds

    def _create(self, k8s_client: K8sApiClient) -> bool:

        apiextensions_v1_api: ApiextensionsV1Api = k8s_client.apiextensions_v1_api
        k8s_object: V1CustomResourceDefinition = self.get_k8s_object()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        try:
            v1_custom_resource_definition: V1CustomResourceDefinition = (
                apiextensions_v1_api.create_custom_resource_definition(
                    body=k8s_object,
                    async_req=self.async_req,
                    pretty=self.pretty,
                )
            )
            # logger.debug("Created: {}".format(v1_custom_resource_definition))
            if v1_custom_resource_definition.metadata.creation_timestamp is not None:
                logger.debug("CustomResourceDefinition Created")
                self.active_resource = v1_custom_resource_definition
                self.active_resource_class = V1CustomResourceDefinition
                return True
        except ValueError as e:
            # This is a K8s bug. Ref: https://github.com/kubernetes-client/python/issues/1022
            logger.warning("Encountered known K8s bug. Exception: {}".format(e))
        logger.error("CustomResourceDefinition could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1CustomResourceDefinition]:
        """Returns the "Active" CustomResourceDefinition from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[V1CustomResourceDefinition] = None
        active_resources: Optional[
            List[V1CustomResourceDefinition]
        ] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_crd.metadata.name: _crd for _crd in active_resources}

        crd_name = self.get_resource_name()
        if crd_name in active_resources_dict:
            active_resource = active_resources_dict[crd_name]
            self.active_resource = active_resource
            self.active_resource_class = V1CustomResourceDefinition
            logger.debug(f"Found active {crd_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:

        apiextensions_v1_api: ApiextensionsV1Api = k8s_client.apiextensions_v1_api
        crd_name = self.get_resource_name()
        k8s_object: V1CustomResourceDefinition = self.get_k8s_object()

        logger.debug("Updating: {}".format(crd_name))
        v1_custom_resource_definition: V1CustomResourceDefinition = (
            apiextensions_v1_api.patch_custom_resource_definition(
                name=crd_name,
                body=k8s_object,
                async_req=self.async_req,
                pretty=self.pretty,
            )
        )
        # logger.debug("Updated: {}".format(v1_custom_resource_definition))
        if v1_custom_resource_definition.metadata.creation_timestamp is not None:
            logger.debug("CustomResourceDefinition Updated")
            self.active_resource = v1_custom_resource_definition
            self.active_resource_class = V1CustomResourceDefinition
            return True
        logger.error("CustomResourceDefinition could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        apiextensions_v1_api: ApiextensionsV1Api = k8s_client.apiextensions_v1_api
        crd_name = self.get_resource_name()

        logger.debug("Deleting: {}".format(crd_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = (
            apiextensions_v1_api.delete_custom_resource_definition(
                name=crd_name,
                async_req=self.async_req,
                pretty=self.pretty,
            )
        )
        # logger.debug("CRD delete_status type: {}".format(type(delete_status.status)))
        # logger.debug("CRD delete_status: {}".format(delete_status.status))
        # TODO: limit this if statement to when delete_status == Success
        if delete_status is not None:
            logger.debug("CustomResourceDefinition Deleted")
            return True
        return False

    def get_k8s_manifest_dict(self) -> Optional[Dict[str, Any]]:
        """Returns the K8s Manifest for a CRD as a dict
        Overwrite this function because the open_apiv3_schema cannot be
        converted to a dict

        Currently we return None meaning CRDs aren't processed by phi k commands
        TODO: fix this
        """
        return None
        # from itertools import chain
        #
        # k8s_manifest: Dict[str, Any] = {}
        # all_attributes: Dict[str, Any] = self.dict(exclude_defaults=True, by_alias=True)
        # # logger.debug("All Attributes: {}".format(all_attributes))
        # for attr_name in chain(
        #     self.fields_for_k8s_manifest_base, self.fields_for_k8s_manifest
        # ):
        #     if attr_name in all_attributes:
        #         if attr_name == "spec":
        #             continue
        #         else:
        #             k8s_manifest[attr_name] = all_attributes[attr_name]
        # # logger.debug(f"k8s_manifest:\n{k8s_manifest}")
        # return k8s_manifest



import base64
from typing import Optional, Dict, List
from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_pod import V1Pod
from kubernetes.stream import stream
from kubernetes.stream.ws_client import WSClient

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.core.v1.pod import Pod
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


def execute_command(
    cmd: List[str],
    pod: Pod,
    k8s_api_client: K8sApiClient,
    container_name: Optional[str] = None,
    wait: bool = True,
    print_output: bool = True,
    k8s_env: Optional[Dict[str, str]] = None,
) -> bool:
    import os
    import socket
    import time

    active_pod: Optional[V1Pod] = pod.read(k8s_api_client)
    # logger.debug("active_pod: {}".format(active_pod.metadata))
    if active_pod is None or not isinstance(active_pod, V1Pod):
        print_error("V1Pod not found")
        return False

    pod_name = active_pod.metadata.name
    print_info(
        "\nRunning command: {}\nPod: {}\nContainer: {}\n".format(
            cmd, pod_name, container_name
        )
    )
    core_v1_api: CoreV1Api = k8s_api_client.core_v1_api

    ws_connection: WSClient = stream(
        core_v1_api.connect_get_namespaced_pod_exec,
        pod_name,
        pod.get_namespace(),
        command=cmd,
        container=container_name,
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
        _preload_content=False,
    )
    # print_info(ws_connection)

    # ws_connection.is_open() returns True if the connection is alive
    while ws_connection.is_open():
        ws_connection.update(timeout=5)
        if ws_connection.peek_stdout():
            op_line: str = ws_connection.read_all()
            print_info(op_line)
    ws_connection.close()

    return True



from phidata.utils.enums import ExtendedEnum


class ApiVersion(ExtendedEnum):
    CORE_V1 = "v1"
    APPS_V1 = "apps/v1"
    RBAC_AUTH_V1 = "rbac.authorization.k8s.io/v1"
    STORAGE_V1 = "storage.k8s.io/v1"
    APIEXTENSIONS_V1 = "apiextensions.k8s.io/v1"
    # CRDs for Traefik
    TRAEFIK_CONTAINO_US_V1ALPHA1 = "traefik.containo.us/v1alpha1"
    CLIENT_AUTHENTICATION_V1ALPHA1 = "client.authentication.k8s.io/v1alpha1"
    CLIENT_AUTHENTICATION_V1BETA1 = "client.authentication.k8s.io/v1beta1"



from pathlib import Path
from typing import Optional, Any

from phidata.utils.log import logger


def get_path_from_workspace_config(
    current_file: Any, add_path_str: Optional[str] = None
) -> Path:
    """
    Used in the config files, this utility function returns the workspace root path
    from within the workspace directory
    If add_path_str is provided, add it to the workspace root path
    Args:
        current_file:
        add_path_str:

    Returns:
        workspace_root_path
    """
    workspace_root_dir: Path = Path(current_file).parent.parent.resolve()
    # logger.debug(f"Workspace root: {workspace_root_dir}")
    if add_path_str:
        return workspace_root_dir.joinpath(add_path_str)
    return workspace_root_dir


def get_relative_path(relative_path_str: str) -> Path:
    """
    A utility function that resolved the relative_path from the current file
    Args:
        relative_path_str:
    """
    return Path(".").joinpath(relative_path_str).resolve()



from typing import Dict, List, Optional
from typing_extensions import Literal

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_resource_definition import (
    CustomResourceDefinition,
    CustomResourceDefinitionSpec,
    CustomResourceDefinitionNames,
    CustomResourceDefinitionVersion,
    V1JSONSchemaProps,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class CreateCustomResourceDefinition(BaseModel):
    crd_name: str
    app_name: str
    group: str
    names: CustomResourceDefinitionNames
    scope: Literal["Cluster", "Namespaced"] = "Namespaced"
    versions: List[CustomResourceDefinitionVersion]
    annotations: Optional[Dict[str, str]] = None
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[CustomResourceDefinition]:
        """Creates a CustomResourceDefinition resource"""

        crd_name = self.crd_name
        logger.debug(f"Init CRD resource: {crd_name}")

        crd_labels = create_component_labels_dict(
            component_name=crd_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        crd_versions: List[CustomResourceDefinitionVersion] = []
        if self.versions is not None and isinstance(self.versions, list):
            for version in self.versions:
                if isinstance(version, CustomResourceDefinitionVersion):
                    crd_versions.append(version)
        else:
            print_error("CustomResourceDefinitionVersion invalid")
            return None

        crd = CustomResourceDefinition(
            api_version=ApiVersion.APIEXTENSIONS_V1,
            kind=Kind.CUSTOMRESOURCEDEFINITION,
            metadata=ObjectMeta(
                name=crd_name,
                labels=crd_labels,
                annotations=self.annotations,
            ),
            spec=CustomResourceDefinitionSpec(
                group=self.group,
                names=self.names,
                scope=self.scope,
                versions=crd_versions,
            ),
        )

        # logger.debug(f"CRD {crd_name}:\n{crd.json(exclude_defaults=True, indent=2)}")
        return crd



from typing import Optional, Literal

from phidata.asset.file import File
from phidata.asset.aws.s3 import S3Object
from phidata.utils.log import logger
from phidata.task import PythonTask, PythonTaskArgs


class DownloadUrlToS3Args(PythonTaskArgs):
    url: str
    s3_object: S3Object
    # How to behave if the object already exists.
    # fail: Raise a ValueError.
    # replace: Replace the object
    # skip: Do nothing
    if_exists: Literal["fail", "replace", "skip"] = "replace"
    # Optional File object to use
    file: Optional[File] = None


class DownloadUrlToS3(PythonTask):
    def __init__(
        self,
        url: str,
        s3_object: S3Object,
        if_exists: Literal["fail", "replace", "append"] = "replace",
        file: Optional[File] = None,
        name: str = "download_url_to_s3",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: DownloadUrlToS3Args = DownloadUrlToS3Args(
                url=url,
                s3_object=s3_object,
                if_exists=if_exists,
                file=file,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=download_url_to_s3,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def url(self) -> str:
        return self.args.url

    @url.setter
    def url(self, url: str) -> None:
        if url is not None:
            self.args.url = url

    @property
    def s3_object(self) -> Optional[S3Object]:
        return self.args.s3_object

    @s3_object.setter
    def s3_object(self, s3_object: S3Object) -> None:
        if s3_object is not None:
            self.args.s3_object = s3_object

    @property
    def if_exists(self) -> Literal["fail", "replace", "skip"]:
        return self.args.if_exists if self.args else "skip"

    @if_exists.setter
    def if_exists(self, if_exists: Literal["fail", "replace", "skip"]) -> None:
        if if_exists is not None:
            self.args.if_exists = if_exists

    @property
    def file(self) -> Optional[File]:
        return self.args.file

    @file.setter
    def file(self, file: File) -> None:
        if file is not None:
            self.args.file = file


def best_guess_file_name(url: str) -> str:

    from phidata.utils.dttm import get_today_utc_date_str

    fn: Optional[str] = None
    if "/" in url:
        fn = url.split("/")[-1]
    if fn is None:
        fn = f"tmp__{get_today_utc_date_str()}"
    logger.debug("fn: {}".format(fn))
    return fn


def download_url_to_s3(**kwargs) -> bool:

    args: DownloadUrlToS3Args = DownloadUrlToS3Args.from_kwargs(kwargs)
    # logger.info(f"DownloadUrlToS3Args: {args}")

    if args.url is None:
        logger.error("URL not available")
        return False

    if args.s3_object is None:
        logger.error("S3Object not available")
        return False

    if args.file is None:
        args.file = File.temporary(best_guess_file_name(args.url))

    download_success = args.file.download_url(args.url)
    if download_success:
        return args.s3_object.upload_file(file=args.file, if_exists=args.if_exists)
    return False



from collections import OrderedDict
from typing import Any, Dict, List, Optional, Type, Set, cast, Tuple

from phidata.infra.aws.args import AwsArgs
from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.exceptions import (
    AwsArgsException,
    AwsApiClientException,
    AwsResourceCreationFailedException,
)
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.infra.aws.resource.types import AwsResourceType
from phidata.infra.aws.resource.utils import (
    get_aws_resources_from_group,
    filter_and_flatten_aws_resource_groups,
)
from phidata.utils.cli_console import (
    print_info,
    print_info,
    print_heading,
    print_subheading,
    print_error,
    confirm_yes_no,
)
from phidata.utils.common import is_empty
from phidata.utils.log import logger


class AwsWorker:
    """This class interacts with the Aws API."""

    def __init__(self, aws_args: AwsArgs) -> None:
        self.aws_args: AwsArgs = aws_args
        self.aws_api_client: AwsApiClient = AwsApiClient(
            aws_region=self.aws_args.aws_region,
            aws_profile=self.aws_args.aws_profile,
        )
        self.aws_resources: Optional[Dict[str, AwsResourceGroup]] = None
        logger.debug(f"**-+-** AwsWorker created")

    def is_client_initialized(self) -> bool:
        return self.aws_api_client.is_initialized()

    def are_resources_initialized(self) -> bool:
        if self.aws_resources is not None and len(self.aws_resources) > 0:
            return True
        return False

    def are_resources_active(self) -> bool:
        # TODO: fix this
        return False

    ######################################################
    ## Init Resources
    ######################################################

    def init_resources(self) -> None:
        """
        This function populates the self.aws_resources dictionary.
        """
        logger.debug("-*- Initializing AwsResourceGroup")

        if self.aws_resources is None:
            self.aws_resources = OrderedDict()

        rgs_to_build = self.aws_args.resources
        if rgs_to_build is not None and isinstance(rgs_to_build, list):
            for resource_group in rgs_to_build:
                self.aws_resources[resource_group.name] = resource_group

        logger.debug(f"# AwsResources built: {len(self.aws_resources)}")

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Creating AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return False

        aws_resources_to_create: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        # Validate resources to be created
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in aws_resources_to_create:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {len(aws_resources_to_create)} resources")
            confirm = confirm_yes_no("\nConfirm deploy")
            if not confirm:
                print_info("Skipping deploy")
                return False

        # track the total number of AwsResources to create for validation
        num_resources_to_create: int = len(aws_resources_to_create)
        num_resources_created: int = 0

        for resource in aws_resources_to_create:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_created = resource.create(aws_client=self.aws_api_client)
                if _resource_created:
                    num_resources_created += 1
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources created: {num_resources_created}/{num_resources_to_create}"
        )
        if num_resources_to_create == num_resources_created:
            return True

        logger.error(
            f"Resources created: {num_resources_created} do not match resources required: {num_resources_to_create}"
        )
        return False

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Creating AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return

        aws_resources_to_create: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        num_resources_to_create: int = len(aws_resources_to_create)
        print_heading("--**-- Aws resources:")
        for resource in aws_resources_to_create:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_create} resources")

    ######################################################
    ## Get Resources
    ######################################################

    def get_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> Optional[List[AwsResource]]:

        logger.debug("-*- Getting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No AwsResources available")
                return None

        aws_resources: List[AwsResource] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        return aws_resources

    def get_resource_groups(self) -> Optional[Dict[str, AwsResourceGroup]]:
        logger.debug("-*- Getting AwsResourceGroups")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No AwsResources available")
                return {}

        return self.aws_resources

    ######################################################
    ## Read Resources
    ######################################################

    def read_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> Optional[List[AwsResource]]:

        logger.debug("-*- Getting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return None

        aws_resources: Optional[
            List[AwsResource]
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )
        return aws_resources

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Deleting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return False

        aws_resources_to_delete: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            sort_order="delete",
        )

        # Validate resources to be deleted
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in aws_resources_to_delete:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {len(aws_resources_to_delete)} resources")
            confirm = confirm_yes_no("\nConfirm delete")
            if not confirm:
                print_info("Skipping delete")
                return False

        # track the total number of AwsResources to delete for validation
        num_resources_to_delete: int = len(aws_resources_to_delete)
        num_resources_deleted: int = 0

        for resource in aws_resources_to_delete:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_deleted = resource.delete(aws_client=self.aws_api_client)
                if _resource_deleted:
                    num_resources_deleted += 1
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be deleted."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources deleted: {num_resources_deleted}/{num_resources_to_delete}"
        )
        if num_resources_to_delete == num_resources_deleted:
            return True

        logger.error(
            f"Resources deleted: {num_resources_deleted} do not match resources required: {num_resources_to_delete}"
        )
        return False

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Deleting AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return

        aws_resources_to_delete: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            sort_order="delete",
        )

        num_resources_to_delete: int = len(aws_resources_to_delete)
        print_heading("--**-- Aws resources:")
        for resource in aws_resources_to_delete:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_delete} resources")

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Patching AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return False

        aws_resources_to_patch: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        # Validate resources to be patched
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in aws_resources_to_patch:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {len(aws_resources_to_patch)} resources")
            confirm = confirm_yes_no("\nConfirm patch")
            if not confirm:
                print_info("Skipping patch")
                return False

        # track the total number of AwsResources to patch for validation
        num_resources_to_patch: int = len(aws_resources_to_patch)
        num_resources_patched: int = 0

        for resource in aws_resources_to_patch:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_patched = resource.update(aws_client=self.aws_api_client)
                if _resource_patched:
                    num_resources_patched += 1
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources patched: {num_resources_patched}/{num_resources_to_patch}"
        )
        if num_resources_to_patch == num_resources_patched:
            return True

        logger.error(
            f"Resources patched: {num_resources_patched} do not match resources required: {num_resources_to_patch}"
        )
        return False

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Patching AwsResources")
        if self.aws_resources is None:
            self.init_resources()
            if self.aws_resources is None:
                print_info("No resources available")
                return

        aws_resources_to_patch: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=self.aws_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        num_resources_to_patch: int = len(aws_resources_to_patch)
        print_heading("--**-- Aws resources:")
        for resource in aws_resources_to_patch:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_patch} resources")

    ######################################################
    ## End
    ######################################################



import os
from typing import Dict, Any, Optional

from phidata.utils.log import logger


def validate_env_vars(env_var_dict: Optional[Dict[str, Any]]) -> bool:
    """
    Return True if each env_var in env_var_dict matches
    its value in the running environment

    Args:
        env_var_dict:

    Returns:
        bool
    """
    # logger.debug("Validating env: {}".format(env_var_dict))
    if env_var_dict is None:
        # nothing to validate
        return True

    if isinstance(env_var_dict, dict):
        # return True if each env_var matches
        for env_var_key, env_var_value in env_var_dict.items():
            _env_value = os.getenv(env_var_key)
            if str(env_var_value) != str(_env_value):
                # logger.debug(f"EnvVar {env_var_key} invalid. Value {_env_value}")
                return False
        return True

    # Return false if env_var_dict is not a dict
    return False



from airflow.models.baseoperator import BaseOperator
from airflow.utils.context import Context


class EmptyOperator(BaseOperator):
    """
    Operator that does literally nothing. It can be used to group tasks in a
    DAG.
    The task is evaluated by the scheduler but never processed by the executor.

    # https://github.com/apache/airflow/blob/main/airflow/operators/empty.py
    # Note: use airflow.operators.empty.EmptyOperator when ^ is merged
    """

    ui_color = "#e8f7e4"
    inherits_from_empty_operator = True

    def execute(self, context: Context):
        pass



from pathlib import Path
from typing import Optional, Dict, Any, List, Union
from typing_extensions import Literal

from phidata.app.db import DbApp
from phidata.infra.k8s.create.core.v1.service import ServiceType
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.app.superset.superset_base import SupersetBase
from phidata.utils.log import logger


class SupersetWebserver(SupersetBase):
    def __init__(
        self,
        name: str = "superset-ws",
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "apache/superset",
        image_tag: str = "latest",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Mount the workspace directory on the container,
        mount_workspace: bool = False,
        workspace_volume_name: Optional[str] = None,
        # Path to mount the workspace volume under,
        # This is the parent directory for the workspace on the container,
        # i.e. the ws is mounted as a subdir in this dir,
        # eg: if ws name is: idata, workspace_dir would be: /usr/local/idata,
        workspace_parent_container_path: str = "/usr/local",
        # NOTE: On DockerContainers the workspace_root_path is mounted to workspace_dir,
        # because we assume that DockerContainers are running locally on the user's machine,
        # On K8sContainers, we load the workspace_dir from git using a git-sync sidecar container,
        create_git_sync_sidecar: bool = True,
        git_sync_repo: Optional[str] = None,
        git_sync_branch: Optional[str] = None,
        git_sync_wait: int = 1,
        # But when running k8s locally, we can mount the workspace using,
        # host path as well.,
        k8s_mount_local_workspace: bool = False,
        # Superset resources directory relative to the workspace_root,
        # This directory contains all the files required by superset.,
        # eg: docker-bootstrap.sh,
        # This dir is mounted to the `/app/docker` directory on the container,
        mount_resources: bool = True,
        resources_volume_name: Optional[str] = None,
        resources_dir: str = "superset",
        resources_dir_container_path: str = "/app/docker",
        # Set the SUPERSET_CONFIG_PATH env var,
        superset_config_path: Optional[str] = None,
        # Set the REQUIREMENTS_LOCAL env var,
        # defaults to "/app/docker/requirements-local.txt",
        requirements_local: Optional[str] = None,
        # Set the PYTHONPATH env var,
        # defaults to "/app/pythonpath",
        python_path: Optional[str] = None,
        # Configure Superset database,
        # Get database details using DbApp,
        db_app: Optional[DbApp] = None,
        # Provide database details,
        # Set the DATABASE_USER env var,
        db_user: Optional[str] = None,
        # Set the DATABASE_PASSWORD env var,
        db_password: Optional[str] = None,
        # Set the DATABASE_DB env var,
        db_schema: Optional[str] = None,
        # Set the DATABASE_HOST env var,
        db_host: Optional[str] = None,
        # Set the DATABASE_PORT env var,
        db_port: Optional[int] = None,
        # Set the DATABASE_DIALECT env var,
        db_dialect: str = "postgresql+psycopg2",
        # Superset db connections in the format { conn_id: conn_url },
        db_connections: Optional[Dict] = None,
        # Configure superset redis,
        # Get redis details using PhidataApp,
        redis_app: Optional[Any] = None,
        # Provide redis details,
        # Set the REDIS_HOST env var,
        redis_host: str = "redis",
        # Set the REDIS_PORT env var,
        redis_port: int = 6379,
        # Set the FLASK_ENV env var,
        flask_env: str = "production",
        # Set the SUPERSET_ENV env var,
        superset_env: str = "production",
        # Set the SUPERSET_LOAD_EXAMPLES env var,
        superset_load_examples: str = "yes",
        # Configure the container,
        container_name: Optional[str] = None,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Add container labels,
        container_labels: Optional[Dict[str, Any]] = None,
        # NOTE: Available only for Docker,
        # Add volumes to DockerContainer,
        # container_volumes is a dictionary which adds the volumes to mount,
        # inside the container. The key is either the host path or a volume name,,
        # and the value is a dictionary with 2 keys:,
        #   bind - The path to mount the volume inside the container,
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.,
        # For example:,
        # {,
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},,
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'},
        # },
        container_volumes: Optional[Dict[str, dict]] = None,
        # Open a container port if open_container_port=True,
        open_container_port: bool = False,
        # Port number on the container,
        container_port: int = 8000,
        # Port name: Only used by the K8sContainer,
        container_port_name: str = "http",
        # Host port: Only used by the DockerContainer,
        container_host_port: int = 8000,
        # Open the app port if open_app_port=True,
        open_app_port: bool = False,
        # App port number on the container,
        app_port: int = 8088,
        # Only used by the K8sContainer,
        app_port_name: str = "app",
        # Only used by the DockerContainer,
        app_host_port: int = 8088,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the deployment,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Add deployment labels,
        deploy_labels: Optional[Dict[str, Any]] = None,
        # Determine how to spread the deployment across a topology,
        # Key to spread the pods across,
        topology_spread_key: Optional[str] = None,
        # The degree to which pods may be unevenly distributed,
        topology_spread_max_skew: Optional[int] = None,
        # How to deal with a pod if it doesn't satisfy the spread constraint.,
        topology_spread_when_unsatisfiable: Optional[
            Literal["DoNotSchedule", "ScheduleAnyway"],
        ] = None,
        # Configure the app service,
        create_app_service: bool = False,
        app_service_name: Optional[str] = None,
        app_service_type: Optional[ServiceType] = None,
        # The port that will be exposed by the service.,
        app_service_port: int = 8088,
        # The node_port that will be exposed by the service if app_service_type = ServiceType.NODE_PORT,
        app_node_port: Optional[int] = None,
        # The app_target_port is the port to access on the pods targeted by the service.,
        # It can be the port number or port name on the pod.,
        app_target_port: Optional[Union[str, int]] = None,
        # Add labels to app service,
        app_service_labels: Optional[Dict[str, Any]] = None,
        # Additional args
        # If True, use cached resources
        # i.e. skip resource creation/deletion if active resources with the same name exist.
        use_cache: bool = True,
    ):
        super().__init__(
            name=name,
            version=version,
            enabled=enabled,
            image_name=image_name,
            image_tag=image_tag,
            entrypoint=entrypoint,
            command=command,
            mount_workspace=mount_workspace,
            workspace_volume_name=workspace_volume_name,
            workspace_parent_container_path=workspace_parent_container_path,
            create_git_sync_sidecar=create_git_sync_sidecar,
            git_sync_repo=git_sync_repo,
            git_sync_branch=git_sync_branch,
            git_sync_wait=git_sync_wait,
            k8s_mount_local_workspace=k8s_mount_local_workspace,
            mount_resources=mount_resources,
            resources_volume_name=resources_volume_name,
            resources_dir=resources_dir,
            resources_dir_container_path=resources_dir_container_path,
            superset_config_path=superset_config_path,
            requirements_local=requirements_local,
            python_path=python_path,
            db_app=db_app,
            db_user=db_user,
            db_password=db_password,
            db_schema=db_schema,
            db_host=db_host,
            db_port=db_port,
            db_dialect=db_dialect,
            db_connections=db_connections,
            redis_app=redis_app,
            redis_host=redis_host,
            redis_port=redis_port,
            flask_env=flask_env,
            superset_env=superset_env,
            superset_load_examples=superset_load_examples,
            container_name=container_name,
            image_pull_policy=image_pull_policy,
            container_detach=container_detach,
            container_auto_remove=container_auto_remove,
            container_remove=container_remove,
            container_labels=container_labels,
            container_volumes=container_volumes,
            open_container_port=open_container_port,
            container_port=container_port,
            container_port_name=container_port_name,
            container_host_port=container_host_port,
            open_app_port=open_app_port,
            app_port=app_port,
            app_port_name=app_port_name,
            app_host_port=app_host_port,
            env=env,
            env_file=env_file,
            config_map_name=config_map_name,
            secret_name=secret_name,
            secrets_file=secrets_file,
            deploy_name=deploy_name,
            pod_name=pod_name,
            replicas=replicas,
            pod_node_selector=pod_node_selector,
            restart_policy=restart_policy,
            termination_grace_period_seconds=termination_grace_period_seconds,
            deploy_labels=deploy_labels,
            topology_spread_key=topology_spread_key,
            topology_spread_max_skew=topology_spread_max_skew,
            topology_spread_when_unsatisfiable=topology_spread_when_unsatisfiable,
            create_app_service=create_app_service,
            app_service_name=app_service_name,
            app_service_type=app_service_type,
            app_service_port=app_service_port,
            app_node_port=app_node_port,
            app_target_port=app_target_port,
            app_service_labels=app_service_labels,
            use_cache=use_cache,
        )
